import importlib
import logging
from pathlib import Path
import yaml


def read_config(path_to_config_file):
    with open(path_to_config_file, 'r') as file:
        tasks = yaml.load(file, Loader=yaml.FullLoader)

    return tasks


def get_task_by_name(
        tasks,
        target_name,
):
    task = None
    for candidate_task in tasks:
        candidate_name = candidate_task.get('name')

        if target_name == candidate_name:
            task = candidate_task

    if task is None:
        task = {}

    return task


def execute_config_task(task):
    output_folder = Path('output')
    output_folder.mkdir(exist_ok=True, parents=True)

    csv_file_name = task.get('csv_file_name', '')
    csv_file_path = output_folder.joinpath(f'{csv_file_name}.csv')

    # Get Source Transport
    path_to_source_transport_module = task.get('source_transport', {}).get('path_to_module', '')
    source_transport_name = task.get('source_transport', {}).get('transport_name', '')
    source_transport_kwargs = task.get('source_transport', {}).get('kwargs', {})

    source_transport_kwargs.update({'csv_file_path': csv_file_path})

    source_transport_module = importlib.import_module(path_to_source_transport_module)
    source_transport = getattr(source_transport_module, source_transport_name)

    # Get Destination Transport
    path_to_destination_transport_module = task.get('destination_transport', {}).get('path_to_module', '')
    destination_transport_name = task.get('destination_transport', {}).get('transport_name', '')
    destination_transport_kwargs = task.get('destination_transport', {}).get('kwargs', {})
    destination_transport_kwargs.update({'csv_file_path': csv_file_path})

    destination_transport_module = importlib.import_module(path_to_destination_transport_module)
    destination_transport = getattr(destination_transport_module, destination_transport_name)

    if source_transport is None:
        raise ValueError('No Source Transport is specified!!!')

    if destination_transport is None:
        raise ValueError('No Destination Transport is specified!!!')


    # Run Source
    source_transport(**source_transport_kwargs).run()
    destination_transport(**destination_transport_kwargs).run()




import abc

class BaseTransport(abc.ABC):

    def __init__(self,
                 csv_file_path,
                 *args,
                 **kwargs
                 ):
        self.csv_file_path = csv_file_path
    @abc.abstractmethod
    def run(self):
        pass



import os
import pandas as pd
import logging
from pathlib import Path


def get_from_env(
        curr_val,
        env_key: str,
        default_val,
):
    if curr_val is None:
        return os.getenv(env_key, default_val)
    else:
        return curr_val


def dump_dataframe_as_csv(
        dataframe: pd.DataFrame,
        csv_file_path,
):

    logging.info('Display dataframe')
    logging.info(dataframe)

    if csv_file_path.exists():
        logging.info(f"deleting old file {csv_file_path}")
        csv_file_path.unlink()

    dataframe.to_csv(csv_file_path, index=False)

    return dataframe



import logging
import os
from typing import Optional, List, Tuple
from pathlib import Path
import psycopg2
from transportlib.utils import get_from_env
from transportlib.transport import BaseTransport
import traceback


class PostgresTransport(BaseTransport):
    """
    Truncate Staging Table
    Load CSV to Staging Table.
    Run Custom SQL which loads from Staging to L0 Table(s).
    """

    def __init__(self,
                 # CSV file path
                 csv_file_path: Path,
                 # Load to staging + L0
                 staging_table_name: str,
                 staging_to_l0_sql: str,
                 staging_to_l0_args: Optional[Tuple],

                 # Credntials
                 host = None,
                 port = None,
                 user = None,
                 password = None,
                 database = None,
                 ):
        super(PostgresTransport, self).__init__(csv_file_path=csv_file_path)
        host = get_from_env(curr_val=host, env_key='POSTGRES_HOST', default_val=None)
        port = get_from_env(curr_val=port, env_key='POSTGRES_PORT', default_val=None)
        user = get_from_env(curr_val=user, env_key='POSTGRES_USER', default_val=None)
        port = get_from_env(curr_val=port, env_key='POSTGRES_PORT', default_val=None)
        database = get_from_env(curr_val=database, env_key='POSTGRES_DATABASE', default_val=None)

        self.conn = psycopg2.connect(host=host, port=port, user=user, password=password, database=database)


        self.cur = self.conn.cursor()
        self.staging_table_name = staging_table_name
        self.staging_to_l0_sql = staging_to_l0_sql
        self.staging_l0_args = staging_to_l0_args

    def run(self):
        # Truncate table
        self.truncate_staging_table()

        # Bulk load data to staging
        try:
            self.load_csv_file_to_staging()

        except Exception as e:
            logging.error("Something went wrong with loading csv data to staging!!!")
            logging.error(traceback.format_exc())
            return False

        # LOad from staging to l0
        try:
            self.load_from_staging_to_l0()
            self.cur.close()
        except Exception as e:
            logging.error("Something went wrong with loading from staging to l0!!!")
            logging.error(traceback.format_exc())
            return False
        finally:
            logging.info("Closing connection....")
            if self.conn is not None:
                self.conn.close()

    def truncate_staging_table(self):
        logging.info(f"Now truncating {self.staging_table_name}...")

        truncate_table_sql = f"""
        TRUNCATE {self.staging_table_name}
        """

        logging.info('Running SQL')
        logging.info(truncate_table_sql)

        self.cur.execute(truncate_table_sql)
        self.conn.commit()

    def load_csv_file_to_staging(
            self,
    ):
        logging.info(f"Now loading file {self.csv_file_path} to staging")
        load_to_staging_sql = f"""
        -- COPY all data from csv into staging table
        
        COPY {self.staging_table_name} FROM stdin with CSV HEADER DELIMITER as ',' QUOTE as '"' ;
        """
        logging.info("COPY JOB")
        logging.info(load_to_staging_sql)

        with open(self.csv_file_path, 'r') as read_file:
            self.cur.copy_expert(sql=load_to_staging_sql, file=read_file)
            self.conn.commit()

    def load_from_staging_to_l0(
            self,
    ):
        logging.info(f"Now loading data from staging to l0 with the below SQL:")
        logging.info(self.staging_to_l0_sql)

        if self.staging_l0_args is None:
            self.cur.execute(query=self.staging_to_l0_sql)
        else:
            self.cur.execute(self.staging_to_l0_sql, self.staging_l0_args)

        # Print notice
        notices = self.conn.notices
        if type(notices) == list:
            for notice in self.conn.notices:
                logging.info(notice)
        else:
            logging.info("No notice is printed")

        self.conn.commit()



import datetime
import logging
import traceback
from pathlib import Path
from typing import List, Optional
import yfinance as yf
import pandas as pd
from transportlib.transport import BaseTransport

from transportlib.utils import dump_dataframe_as_csv


class HistoricalStockPriceTransport(BaseTransport):

    def __init__(
            self,
            csv_file_path: Path,
            start_datetime: datetime.datetime,
            end_datetime: datetime.datetime,
            tickers: Optional[List[str]] = None,
    ):
        super(HistoricalStockPriceTransport, self).__init__(csv_file_path=csv_file_path)

        self.start_datetime = start_datetime
        self.end_datetime = end_datetime

        if tickers is None:
            logging.info("Tickers set to None. Running all available HK tickers.")
            tickers = [f"{str(i).rjust(4, '0')}.hk" for i in range(1, 9999)]

        self.tickers = tickers

    def run(self):
        hist_df = pd.DataFrame()

        try:
            for ticker in self.tickers:
                logging.info(f'running ticker {ticker}')

                hist_df_new: pd.DataFrame = yf.download(
                    tickers=ticker,
                    start=self.start_datetime,
                    end=self.end_datetime,
                )

                hist_df_new = hist_df_new.reset_index()
                hist_df_new.insert(0, 'ticker', ticker)
                hist_df_new = hist_df_new.rename(
                    columns=
                    {
                        'Date': 'date',
                        'Open': 'open',
                        'High': 'high',
                        'Low': 'low',
                        'Close': 'close',
                        'Adj Close': 'adj_close',
                        'Volume': 'volume',
                    }
                )
                hist_df_new['lib'] = 'yfinance'
                hist_df_new['lib'] = 'yfinance'
                hist_df_new['lib_called_at'] = datetime.datetime.utcnow()

                hist_df = pd.concat([hist_df, hist_df_new])
        except KeyboardInterrupt as e:
            logging.error(traceback.format_exc())


        path_to_csv_file = dump_dataframe_as_csv(dataframe=hist_df, csv_file_path=self.csv_file_path)

        return path_to_csv_file
